from .. import util
import numpy as np
from sklearn.cluster import KMeans


def get_neighborhood(columns, centerpoints,k=20):
    surr_cols = util.point.get_nearest_points(columns,centerpoints,k=k)[0]
    rel = (np.swapaxes(surr_cols,0,1) - surr_cols[:,0,:]).reshape(-1,2)
    return rel

def cluster_neighbors(neighborhood, n_clusters=40):
    kmeans = KMeans(n_clusters=n_clusters,copy_x=True,random_state=1337,n_init='auto').fit(neighborhood) # (n,2)
    return kmeans.cluster_centers_

def get_vector_to_neighbors(columns,origins,guess_vector, threshold=5):
    match_col,match_idx,dist = util.point.get_nearest_points(columns,origins+guess_vector,k=1)
    threshold_mask = dist > threshold
    match_col[threshold_mask,:] = np.nan
    return match_col - origins


import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import center_of_mass
from .. import util


'''

'''
def prepare_fourier_pattern(image,log=False):
    if log:
        return np.log(1e2+np.abs(np.fft.fftshift(np.fft.fft2(image))))
    else:
        return np.fft.fftshift(np.fft.fft2(image))

'''

'''
# TODO: delete_within -- arg where if clicked pt is within this of another pt, delete that pt instead of adding
def select_peaks(pattern,cmap='gray',vmin=None,vmax=None,zoom=None,figsize=None,select_conjugates=False,delete_within=None):

    x = []
    y = []
    if figsize is None:
        fig,ax = plt.subplots(1,1,constrained_layout=True)
    else:
        fig,ax = plt.subplots(1,1,constrained_layout=True,figsize=figsize)


    def replot():
        ax.matshow(np.real(pattern),vmin=vmin,vmax=vmax,cmap=cmap)
        if zoom is not None:
            ax.set_xlim(pattern.shape[0]/2 - zoom ,pattern.shape[0]/2 + zoom)
            ax.set_ylim(pattern.shape[1]/2 + zoom ,pattern.shape[1]/2 - zoom)
        ax.axis('off')

    def _onclick_event(event):
        new_x = event.xdata
        new_y = event.ydata
        added_point = False
        if delete_within is not None and len(x) > 0:
            _,idx,dist = util.point.get_nearest_points(np.array((y,x)).T,(new_y,new_x),k=1)
            if dist < delete_within:
                del x[idx]
                del y[idx]
                ax.clear()
                #ax.matshow(np.real(pattern),vmin=vmin,vmax=vmax,cmap=cmap)
                replot()
                ax.plot(x,y,'r.')
            else:
                x.append(new_x)
                y.append(new_y)
                ax.plot(new_x,new_y,'r.')
                added_point = True
        else:
            x.append(new_x)
            y.append(new_y)
            ax.plot(new_x,new_y,'r.')
            added_point = True
        # TODO: test works properly for non square
        # TODO: test off by one for odd / even with flooring...
        if select_conjugates and added_point:
            size_x = pattern.shape[1]
            size_y = pattern.shape[0]
            x_conj = -(new_x-(size_x//2.)) + (size_x//2.)
            y_conj = -(new_y-(size_y//2.)) + (size_y//2.)
            x.append(x_conj)
            y.append(y_conj)
            ax.plot(x_conj,y_conj,'r.')
    fig.canvas.mpl_connect('button_press_event',_onclick_event)
    replot()
    return (y,x)
    #return np.stack((x,y))

# TODO: lots. fit, better viz, better store, check impl.
def refine_peaks_com(pattern, p0, crop_window,iters=10,viz=False):
    p_ref = np.zeros_like(p0) #np.zeros((len(p0[0]),2)) # (n,2)
    for it,pt in enumerate(p0):
        y0,x0 = pt
        for refnum in range(iters):
            crop = util.general.normalize(pattern[int(y0) - crop_window: int(y0) + crop_window, int(x0) - crop_window: int(x0) + crop_window])
            yr,xr = center_of_mass(crop)
            x0 = xr - crop_window + x0
            y0 = yr - crop_window + y0

        p_ref[it,:] = (y0,x0)

        if viz:
            fig,ax = plt.subplots(1,1)
            ax.matshow(crop,cmap='gray')
            ax.plot(xr,yr,'rx')
            ax.plot(crop_window,crop_window,'bo')
    return p_ref
def refine_peaks_gf(pattern, p0, window_dimension=5,store_fits=True, remove_unfit = True):
    return util.general.gaussian_fit_peaks(pattern, p0, window_dimension=window_dimension,store_fits=store_fits, remove_unfit = remove_unfit)


import numpy as np
from .. import util

# TODO: test with non-square
def create_gaussian_filter(shape,p,sigma):
	p = np.array(p)
	filt = np.zeros(shape)
	YY,XX = np.meshgrid(np.arange(shape[0]),np.arange(shape[1]),indexing='ij')
	if len(p.shape)==2:
		for pt in np.array(p).T:
			filt = filt + np.reshape(util.func.gaussian_2d((YY,XX),1,pt[1],pt[0],sigma,sigma,0,0),shape)
	else:
		filt = np.reshape(util.func.gaussian_2d((YY,XX),1,p[1],p[0],sigma,sigma,0,0),shape)
	return filt

def fourier_filter(pattern_c,p,sigma):
	filt = create_gaussian_filter(pattern_c.shape,p,sigma)
	filtered_pattern_c = pattern_c * filt
	filtered_im_c = np.fft.ifft2(np.fft.ifftshift(filtered_pattern_c))

	return filtered_im_c, filtered_pattern_c, filt

def coarsening_length(size,sigma,c=4):
	return size / (2*np.pi*sigma) * np.sqrt(np.log(c))

def damp_complex_amplitude(array_c,level):
    return array_c * level / np.abs(array_c)

def mask_peaks_circle(pattern_c,p,levels,radius):
    masked_pattern_c = pattern_c.copy()
    YY,XX = np.meshgrid(np.arange(pattern_c.shape[0]),np.arange(pattern_c.shape[1]),indexing='ij')
    for it,pt in enumerate(p):
        circle_mask = ((XX-pt[1])**2 + (YY-pt[0])**2 )< radius**2
        masked_pattern_c[circle_mask] = damp_complex_amplitude(masked_pattern_c[circle_mask], levels[it])
    masked_image = np.abs(np.fft.ifft2(np.fft.ifftshift(masked_pattern_c)))
    return masked_pattern_c, masked_image



import numpy as np
from .. import util
import scipy.optimize as opt
from concurrent.futures import ProcessPoolExecutor


def create_patches(grating, patch_size, step_size):
    '''
        path_size must be odd
    '''
    assert patch_size % 2 != 0
    patch_radius = patch_size // 2
    patches = []
    Ypoints = np.arange(grating.shape[0])[patch_radius:-(patch_radius+1):step_size]
    Xpoints = np.arange(grating.shape[1])[patch_radius:-(patch_radius+1):step_size]
    for yy in Ypoints:
        for xx in Xpoints:
            patches.append(grating[yy-patch_radius:yy+patch_radius+1,xx-patch_radius:xx+patch_radius+1])
    patches = np.array(patches)
    shape = (len(Ypoints),len(Xpoints))
    return patches,Ypoints,Xpoints,shape


def fit_patch(args):
    '''
        args:
            0 patch
            1 guess
    '''
    patch = args[0]
    guess = args[1]
    yx = np.meshgrid(np.arange(patch.shape[0]),np.arange(patch.shape[1]))

    try: 
        popt,pcov = opt.curve_fit(util.func.sin2D, yx,patch.ravel(),
            p0 = guess, method='lm', ftol=1e-12,
            xtol=1e-12,gtol=1e-12,maxfev=4000)
        perr = np.sqrt(np.diag(pcov))
    except RuntimeError:
        popt = initial_guess
        perr = (np.nan,np.nan,np.nan,np.nan)

    data_fits = np.stack((patch,util.func.sin2D(yx,*popt).reshape(patch.shape)),axis=2)

    return popt,perr,data_fits

def fit_grating(grating,patch_size,step_size,guess,chunksize=None,verbose=True):
    assert grating.shape[0] == grating.shape[1]
    assert len(grating.shape)==2

    patches,Ypoints,Xpoints,subsample_shape = create_patches(grating,patch_size,step_size)
    sampled_points = np.array(np.meshgrid(Ypoints,Xpoints,indexing='ij')).T.reshape(-1,2)
    npatches = patches.shape[0]
    if verbose:
        print(f'# of patches: {npatches}')
    if chunksize == None:
        chunksize = npatches//20
        if verbose:
            print(f'Using chunk size: {chunksize}')

    executor = ProcessPoolExecutor()
    futures = executor.map(fit_patch,[(patch/grating.max(),guess) for patch in patches],chunksize=chunksize)
    executor.shutdown()

    opts = np.zeros((npatches,4))
    errs = np.zeros((npatches,4))
    for it,future in enumerate(futures):
        opts[it,:] = future[0]
        errs[it,:] = future[1]

    amplitude = np.abs(opts[:,0].reshape(subsample_shape))
    rotation = opts[:,2].reshape(subsample_shape)
    spacing = 2*np.pi / opts[:,1].reshape(subsample_shape)

    return amplitude,spacing,rotation,sampled_points


def peak_to_fit_guess(pt,imshape):
    c = np.array(imshape//2)
    r = np.linalg.norm(pt-c)
    r= 2*np.pi *r / imshape[0]
    theta = np.arctan2((pt-c)[1],(pt-c)[0])
    return (1,r,theta,0)

def test_fit(grating,patch_size,step_size,guess,test_patch_idx = None):
    patches,Ypoints,Xpoints,subsample_shape = create_patches(grating,patch_size,step_size)
    npatches = patches.shape[0]
    if test_patch_idx is None:
        test_patch_idx = npatches//4
    return fit_patch((patches[test_patch_idx,:,:],guess))




import numpy as np
from scipy.signal import medfilt2d


def phase_to_strain(phase, qx, qy, mask_threshold=2.):
    grad_phi = np.gradient(phase)

    # Normalized such that a shift of 2pi over 1 wavelength gives eps=1
    eps_compression = -(grad_phi[0]*qy + grad_phi[1]*qx)/(qx**2+qy**2)*phase.shape[0]/(2*np.pi)
    eps_shear = -(0.5)*(-grad_phi[0]*qx + grad_phi[1]*qy)/(qx**2+qy**2)*phase.shape[0]/(2*np.pi)

    mask = np.abs(eps_shear)>mask_threshold
    eps_temp = np.where(mask, 0, eps_shear)
    for i in range(3):
        eps_temp = medfilt2d(eps_temp, kernel_size=5)
        eps_shear = np.where(mask, eps_temp, eps_shear)


    mask = np.abs(eps_compression)>mask_threshold
    eps_temp = np.where(mask, 0, eps_compression)
    for i in range(3):
        eps_temp = medfilt2d(eps_temp, kernel_size=5)
        eps_compression = np.where(mask, eps_temp, eps_compression)
    return eps_compression, eps_shear


from .. import util
import numpy as np
import skimage
#from scipy.ndimage import gaussian_filter, label,center_of_mass
import matplotlib.pyplot as plt

# based off stemtool
def find_columns(image, distance = 10, threshold = 0.1, deduplicate = False):
    image_norm = image - image.min()
    image_norm = image_norm / image_norm.max()
    thresh_mask = image_norm > threshold
    masked = (image_norm * thresh_mask) - threshold # max 1-threshold, min 0
    masked = masked / masked.max()
    peaks = skimage.feature.peak_local_max(masked,min_distance = distance)
    #peaks_mask = np.zeros_like(image,dtype=bool)
    #peaks_mask[tuple(peaks.T)] = True
    #peak_labels = label(peaks_mask)[0]
    #merged_peaks = center_of_mass(peaks_mask,peak_labels,range(1,np.max(peak_labels)+1))
    
    #peaks = np.array(merged_peaks)
    return peaks


def refine_columns(image, columns0, window_dimension=5,store_fits=True, remove_unfit = True):
    return util.general.gaussian_fit_peaks(image, columns0, window_dimension=window_dimension,store_fits=store_fits, remove_unfit = remove_unfit)


import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap, Normalize
from matplotlib.patches import Polygon, Circle
from matplotlib.collections import LineCollection,PatchCollection

import cmocean
from skimage.util import montage

def plot_numbered_points(image,points,ax=None,delta=0,delta_step=0,verbose=0,zoom=100,vmin=None,vmax=None,color='r'):
    x = np.array(points)[:,1]
    y = np.array(points)[:,0]
    if ax is None:
    	fig,ax = plt.subplots(1,1,constrained_layout=True)
    if image is not None:
    	ax.matshow(image,cmap='gray',vmin=vmin,vmax=vmax)
    	ax.set_xlim(image.shape[0]/2 - zoom ,image.shape[0]/2 + zoom)
    	ax.set_ylim(image.shape[1]/2 + zoom ,image.shape[1]/2 - zoom)

    ax.plot(x,y,'.',color=color)
    for it in range(len(x)):
        if verbose:
            print(str(it)+':'+str(x[it])+','+str(y[it]))
        ax.text(x[it]+delta,y[it]+delta,str(it),color='w',fontsize=4)
        delta = delta+delta_step


def plot_fit_comparison(data_fit_stack,figsize=(6,3),cmap='viridis'):
    if len(data_fit_stack.shape)==3:
        data_fit_stack = np.expand_dims(data_fit_stack,2)
    fig,ax = plt.subplots(1,2,constrained_layout=True,figsize=figsize)
    ax[0].matshow(montage(np.rollaxis(data_fit_stack[:,:,:,0],2,0),padding_width=10,fill=0),cmap=cmap)
    ax[0].set_title('Data')
    ax[0].axis('off')
    ax[1].matshow(montage(np.rollaxis(data_fit_stack[:,:,:,1],2,0),padding_width=10,fill=0),cmap=cmap)
    ax[1].set_title('Fit')
    ax[1].axis('off')
    return fig,ax

def neighborhood_scatter_plot(neighborhood,ax=None,clusters=None):
    if ax is None:
        fig,ax = plt.subplots(1,1,constrained_layout=True)
    ax.axhline(y=0,color='k')
    ax.axvline(x=0,color='k')
    ax.plot(neighborhood[:,1],neighborhood[:,0],'.',alpha=.5,markersize=1)
    if clusters is not None:
        ax.plot(clusters[:,1],clusters[:,0],'.')
        for it, ctr in enumerate(clusters):
            ax.text(ctr[1]+2,ctr[0],str(it),color='red')
    ax.invert_yaxis()
    return ax


def plot_scalar_sites(ax,scalar, sites,s=5,cmap='inferno',**scatterargs):
    mappable=ax.scatter(x = sites[:,1],y=sites[:,0],s=s,c=scalar,cmap=cmap,**scatterargs)
    return mappable

def plot_scalar_bonds(ax,scalar,sites_1,sites_2,cmap='inferno',linewidth=1,vmin=None,vmax=None,**lcargs):
    vnorm = Normalize(vmin=vmin,vmax=vmax,clip=False)
    segs = np.stack((sites_1[:,::-1],sites_2[:,::-1]),axis=1)
    line_segments = LineCollection(segs, array=scalar,norm = vnorm,
                                   linewidths=(linewidth),
                                   linestyles='solid',cmap=cmap,**lcargs)
    ax.add_collection(line_segments)
    return line_segments


def plot_phase(phase,ax=None):
    if ax is None:
        fig,ax = plt.subplots(1,1,constrained_layout=True)
    ctours = [-np.sqrt(3)/2,0,np.sqrt(3)/2]
    mappable = ax.matshow(phase,cmap=cmocean.cm.phase,alpha=1,vmin=0,vmax=2*np.pi)
    ctours = [-np.sqrt(3)/2,0,np.sqrt(3)/2]#np.arange(0,2*np.pi,np.pi/4)
    ax.contour(np.sin(phase),ctours,colors='black',alpha=1, linewidths=1,linestyles='solid')
    ax.axis('off')
    return ax, mappable

def plot_displaced_site(columns,displacements,scale,colors='angle',ax=None,cmap='hsv',linewidth=.2,shape=4,angleshift=0,disp_min=0,disp_max=np.inf,scale_power=0.5):
    mask_sites = (np.linalg.norm(displacements,axis=1) < disp_max) & (np.linalg.norm(displacements,axis=1) > disp_min)
    x0 = columns[mask_sites,1] 
    y0 = columns[mask_sites,0] 
    dx0 = displacements[mask_sites,1]* scale
    dy0 = displacements[mask_sites,0]* scale
    angles = np.arctan2(dy0,dx0)
    patches=[]
    for i in range(len(x0)):
        y,x=y0[i],x0[i]
        dy,dx=dy0[i],dx0[i]
        L=(dx**2.+dy**2.)**(.5)
        L = L ** scale_power
        #L=L**.5 
        L2=L/shape
        xy = np.array([[x,y],[x,y],[x,y]]) +np.array([[L*np.cos(angles[i]), L*np.sin(angles[i])], [L2*np.sin(angles[i]), -L2*np.cos(angles[i])], [-L2*np.sin(angles[i]), L2*np.cos(angles[i])]])
        triangle=Polygon(xy, closed=True)
        patches.append(triangle)
    p = PatchCollection(patches, cmap=cmap, alpha=1)
    
    if type(colors) is str and colors=='angle':
        p.set_array((angles+angleshift)%(2*np.pi))
    elif type(colors) is str and colors=='mag':
        p.set_array(np.linalg.norm(displacements,axis=1))
    elif type(colors) is np.ndarray and colors.shape == angles.shape:
        p.set_array(np.array(colors))
    else:
        print('No usable colors')
    p.set_edgecolor('k')
    p.set_linewidth(linewidth)
    if ax is None:
        fig,ax = plt.subplots(1,1,constrained_layout=True)
    cax=ax.add_collection(p)
    return cax


def coarsening_marker(axis,coarsening_radius,position_frac=(0.95,.95),**kwargs):
    posx = position_frac[0]*axis.get_xlim()[1]
    posy = position_frac[1]*axis.get_ylim()[0]
    coarsening_marker = Circle((posx,posy),radius=coarsening_radius,**kwargs)
    axis.add_patch(coarsening_marker)
    return coarsening_marker




import numpy as np
from tqdm import tqdm
from scipy.optimize import curve_fit
from scipy import interpolate
from . import func
### normalization
#min=0,max=1, no type conversions
def normalize(data):
    data = data - np.min(data)
    data = data / np.max(data)
    return data
def normalize_sum(data):
    return data/data.sum()
def normalize_max(data):
    return data/data.max()



def gaussian_fit_peaks(image, peaks0, window_dimension=5,store_fits=True, remove_unfit = True):
    '''
        window_dimension must be odd
    '''
    assert window_dimension % 2 != 0
    if len(peaks0.shape) == 1:
        peaks0 = np.expand_dims(peaks0,axis=0)

    winrad = window_dimension // 2
    
    x0 = peaks0[:,1]
    y0 = peaks0[:,0]
    n_sites = x0.shape[0]
    xf = np.zeros(x0.shape)
    yf = np.zeros(y0.shape)
    errors = np.zeros(x0.shape,dtype=bool)
    opts = np.zeros((n_sites,7))
    data_fits = np.zeros((window_dimension,window_dimension,n_sites,2))
    
    YY,XX = np.meshgrid(np.arange(-winrad,winrad+1),np.arange(-winrad,winrad+1),indexing='ij')
    for it in tqdm(range(n_sites)):
        x0_i = int(x0[it])
        y0_i = int(y0[it])
        
        ydata = image[y0_i - winrad : y0_i + winrad + 1, x0_i - winrad : x0_i + winrad + 1]
        if ydata.shape != (window_dimension, window_dimension):
            errors[it] = True
            continue
        
        bounds = [ (0,-winrad,-winrad,0,0,0,-np.inf),
                   (np.inf,winrad,winrad,window_dimension,window_dimension,2*np.pi,np.inf)]
        initial_guess = (ydata[winrad, winrad], 0, 0, winrad*.8, winrad*.8, 0, 0)

        try:
            popt,pcov = curve_fit(func.gaussian_2d,(YY,XX), ydata.flatten(),p0=initial_guess,bounds=bounds)
            xf[it] = popt[1]+float(x0_i)
            yf[it] = popt[2]+float(y0_i)
            opts[it,:] = popt
            data_fits[:,:,it,0] = ydata
            data_fits[:,:,it,1] = func.gaussian_2d((YY,XX),*popt).reshape(XX.shape)
            
        except RuntimeError:
            errors[it] = True
            
    if errors.sum() > 0:
        if remove_unfit:
            print(f'Errors with indices: {np.where(errors)[0]}, removed')
            xf = np.delete(xf,errors)
            yf = np.delete(yf,errors)
            opts = np.delete(opts,errors,axis=0)
            data_fits = np.delete(data_fits,errors,axis=2)
        else:
            print(f'Errors with indices: {np.where(errors)[0]}, set to NaN')
            xf[errors] = x0[errors]
            yf[errors] = y0[errors]

    return np.array((yf,xf)).T, errors, opts, data_fits


def rasterize_from_points(points,values,output_shape,method='nearest',fill_value=np.nan):
    xi = np.array(np.meshgrid(np.arange(output_shape[0]),np.arange(output_shape[1]),indexing='ij')).T
    interp = interpolate.griddata(points,values.ravel(),xi,method=method,fill_value=fill_value)
    return interp    





import numpy as np
from scipy.spatial import KDTree


def get_nearest_points(points,guesses,k=1):
    kd = KDTree(points)
    dist,idx = kd.query(guesses,k=k)
    # check that these make sense for non zero k.......
    match_points = points[idx]
    return (match_points,idx,dist)


from .. import util
import numpy as np

def measure_displacements(reference_columns,distorted_columns,threshold=2):
    corresp_dist_cols,_,dists = util.point.get_nearest_points(distorted_columns,reference_columns,k=1)
    disps = corresp_dist_cols - reference_columns
    disps[dists > threshold] = np.nan
    return disps



import numpy as np
from scipy.ndimage import gaussian_filter

def faux_highpass_filter(image,lowpass_sigma):
	return image - gaussian_filter(image,lowpass_sigma)


def hann_filter(image):
    nx,ny = image.shape
    rx,ry = np.meshgrid(np.arange(nx),np.arange(ny))
    rx = rx.T
    ry = ry.T
    mask_realspace = (np.sin(np.pi*rx/nx)*np.sin(np.pi*(ry/ny)))**2
    fim = image*mask_realspace
    fft = np.fft.fft2(fim)
    return (fim, fft)

# www.roberthovden.com/tutorial/2015_fftartifacts.html
def periodic_plus_smooth_decomposition(image):
    [rows,cols] = np.shape(im) 
    #Compute boundary conditions 
    s = np.zeros( np.shape(im) ) 
    s[0,0:] = im[0,0:] - im[rows-1,0:] 
    s[rows-1,0:] = -s[0,0:] 
    s[0:,0] = s[0:,0] + im[0:,0] - im[:,cols-1] 
    s[0:,cols-1] = s[0:,cols-1] - im[0:,0] + im[:,cols-1] 
    #Create grid for computing Poisson solution 
    [cx, cy] = np.meshgrid(2*np.pi*np.arange(0,cols)/cols, 2*np.pi*np.arange(0,rows)/rows) 

    #Generate smooth component from Poisson Eq with boundary condition 
    D = (2*(2 - np.cos(cx) - np.cos(cy))) 
    D[0,0] = np.inf # Enforce 0 mean & handle div by zero # -?
    S = np.fft.fft2(s)/D 
    #S[0,0] = 0 # +?

    P = np.fft.fft2(im) - S # FFT of periodic component 


    return(P,S)
